

;*********** [ Blei-Akku-Lader fr 12 und 24 Volt  ********************
;
;   Datei 		:	sla.asm
;   Autor		:	Hanno Grff
;   Version		:	1.4
;   Datum		:	Februar 2010
;
;   Compiler    :	Atmel AVR Assembler
;
;	nderungen  :	1.1 Tastenentprellung nach P. Dannegger
;				 	1.2 Strom auf 10 A erhht, Timer0-Prescaler von 1024 auf 256 gendert	
;					1.3	Korrekurfaktor fr LCD-Strommessung und Grenzwert bei Uo-Ladung
;					1.4	Softstart
;	
;	Bemerkungen:	
;
;
;********************************************************************************************


; Include-Files
.include "tn26def.inc"
.include "def.inc"
.include "sla.inc"


.dseg
	.org RAMBEGIN

 	Grenzwert:		.byte 1		; Grenzwert fr Uo-Abschaltung bei 1/5 bzw 1/10 Kapazitt
 	Puffer:			.byte 8		; Ringpuffer
	Ladestrom:		.byte 2		



.cseg
	.org 0
	 rjmp POWERON     			;Reset handler
	 rjmp nix ;EXT_INT0			;IRQ0 handler
	 rjmp nix ;PIN-Change
	 rjmp nix ;TIM1_CMP1A  		;Timer1 compare match
	 rjmp nix ;TIM1_CMP1B		;Timer1 compare Match
	 rjmp nix ;TIM1_OVF     	;Timer1 overflow handler
	 rjmp TIM0_OVF     			;Timer0 overflow handler
	 rjmp nix ;USI_Start  		;USI Start handler
	 rjmp nix ;USI_OVF	   	    ;USI Overflow handler
	 rjmp nix ;EE_RDY 	   		;EEPROM Ready handler
	 rjmp nix ;ANA_COMP     	;Analog Comparator handler
	 rjmp nix ;ADC_Complete		;
nix:
	 reti


txtAkku:
 	.db  "Akku 12V:",0				;10 Bytes
 	.db  "Akku 24V:",0

txtMode:
 	.db  " 1. Io",0,0				;8 Bytes! mu gerade Zahl sein!
 	.db  " 2. Uo",0,0				;
 	.db  " 3. Ut",0,0

txtAmpere:
 	.db  " 2,0 A",0,0				;8 Bytes!
 	.db  " 2,5 A",0,0
 	.db  " 3,0 A",0,0
 	.db  " 3,5 A",0,0
 	.db  " 4,0 A",0,0
 	.db  " 4,5 A",0,0
 	.db  " 5,0 A",0,0
 	.db  " 5,5 A",0,0
 	.db  " 6,0 A",0,0
 	.db  " 6,5 A",0,0
 	.db  " 7,0 A",0,0
 	.db  " 7,5 A",0,0
 	.db  " 8,0 A",0,0
	.db  " 8,5 A",0,0
	.db  " 9,0 A",0,0
	.db	 " 9,5 A",0,0
	.db  "10,0 A",0,0

txtSonstiges:
	.db	 "       ",0
	.db	 "  Ende!",0


PowerOn:
	cli									; Interrupts ausschalten
	ldi		R16, RAMEND					; Stack initialisieren
	out	 	SP,  R16

	
	;PINB1 (PWM), lcd_enable und LCD_RS = Ausgang
	ldi 	UL, (1<< PINB1) | (1<<PIN_LCD_ENABLE) | (1<<PIN_LCD_RS)
 	out 	DDRB, UL
	
	;PIN_LCDs (Daten) = Ausgang
	ldi 	UL, (1<<PIN_LCD_D4) | (1<<PIN_LCD_D5) | (1<<PIN_LCD_D6) | (1<<PIN_LCD_D7)
 	out 	DDRA, UL

	ldi		UL, (1 << PLLE)				; PLL aktivieren d.h. asyncroner Takt bei PWM
	out		PLLCSR, UL					; 1 Mhz * 64 = 64 Mhz


waitpll:								; Takt setzen
	in 		UL, PLLCSR
	sbrs	UL, PLOCK
	rjmp	waitpll

	in 		UL,PLLCSR
	ldi		r17, (1 << PCKE)
	or		UL, r17
	OUT		PLLCSR, UL
	
		
	;LCD initialisieren
	rcall	lcd_init
	rcall 	lcd_clear
	
	
	; Spannungen in Register einlesen
	ldi		UL, low(V_Charge_fast)		; Ladeschluspannung (12V) einlesen
	mov		FAST_V_L,UL					; Wert in Registern speichern
	ldi		UL, high(V_Charge_fast)
	mov		FAST_V_H,UL

	ldi		UL, low(V_Charge_Trickle)	; Erhaltungsladespannung einlesen
	mov		TRICKLE_V_L, UL				; in Registern speichern
	ldi		UL, high(V_Charge_Trickle)	
	mov		TRICKLE_V_H, UL				

	;Timer initialisieren	
	ldi		UL, (1<<CS02) 				; Timer/Counter0, prescaler auf 1/256
	out		TCCR0, UL
	ldi		UL, (1<<TOIE0)				; Timer-Overflow0 setzen
	out		TIMSK,UL					; 
	

	;PWM-initialisieren
	ldi		UL, 159				    	; OCR1C setzen 159 = 25khz/50khz
	out		OCR1C, UL
	ldi		UL, 10						; unten anfangen
	out		OCR1A,UL


	;"Akku:" ausgeben				
	ldi   	ZL,Low(txtAkku*2)  			; Label "Akku 12 V: "
 	ldi   	ZH,High(txtAkku*2)
 	rcall	lcd_home
	rcall 	lcd_print
	
	ldi   	ZL,Low(txtAmpere*2)  		; Label "Stromladen: "
 	ldi   	ZH,High(txtAmpere*2)
	ldi		AL,0xA						; 1.Zeile 11. Stelle
	rcall	lcd_cursor					; Cursor setzen
	rcall 	lcd_print


	ser		UL							; Flags setzen
	clr		UH
	clr		Key_Old
	clr		Key_State
	clr		Key_Press

	sei									; Interrupts ermglichen

	
;**************************************************************************************************************
; Wenn Taste 1 gedrckt, dann wenn Akku an ist, starten
; Wenn Taste 1 gedrckt und kein Akku an, dann Volt hochsetzen
; Wenn Taste 2 gedrckt, Strom hochsetzen
;**************************************************************************************************************

start:	
	push	UL
	push	UH
	rcall   get_voltage					; Spannung messen	
	pop		UH
	pop		UL
	ldi		AL,100						; willkrlicher Startwert, ob eine Spannung
	clr		AH	
	cp		AV_L,AL						; gemessen werden kann <=> 3 Volt
	cpc		AV_H,AH
	brsh	akku_on
	rcall	print_voltage
	
	cli									; Interrupts ausschalten
	sbrc	Key_Press, Taste1			; Taste1 gedrckt?
	rcall 	voltage						; Akku-Spannung einstellen
	sbrc	Key_press, Taste2
	rcall	strom						; Akku-Strom einstellen
	sei									; Interrupts wieder zulassen
	rjmp	start						; zurck
	
; In erster LCD-Zeile Akkutyp 12 od. 24 V anzeigen
voltage:
	clr		Key_press					; Taste lschen
	rcall	lcd_home					; Cursor setzen
	ldi   	ZL,Low(txtAkku*2)  			; Label "Akku 12 V: " adressieren
	ldi   	ZH,High(txtAkku*2)
	tst		UL							; Ist Flag = 0
	breq	v1							; dann 12 Volt ausgeben
	adiw	ZH:ZL,10					; sonst Adressierung auf 24 V
v1:	rcall 	lcd_print					; auf LCD ausgeben
	com		UL							; Flag 'umdrehen'
	ret

; und gewhlten Ladestrom
strom:
	clr		key_press					; Taste lschen
	ldi		AL,0xA						; 1.Zeile 11. Stelle
	rcall	lcd_cursor					; Cursor setzen
	ldi   	ZL,Low(txtAmpere*2)  		; Label "Stromladen: "
 	ldi   	ZH,High(txtAmpere*2)
	
	cpi		UH,16						; Ist Max erreicht (=10,0 Amp) ?
	brlo	s0							; wenn nein, weiter bei s0
	clr		UH							; wenn ja, Flag = 0
	rjmp	s3							; Wert ausgeben
	
s0:	inc		UH							; sonst Tastendruck UH=UH+1

    mov		AL,UH
s2: adiw	ZH:ZL,8						; pro Zhlwert in UH
	dec		AL							; zur Speicherstelle
	brne	s2							; 8 Bytes addieren

s3:	rcall	lcd_print					; Ausgabe
	ret


;  Akku angeschaltet, d. h. auf Start warten
akku_on:
	push	UL
	push	UH
	rcall	get_voltage					; Spannung messen	
	pop		UH
	pop		UL
	rcall	print_voltage				; Spannung ausgeben
	ldi		AL,100						; ist Spannung immer noch grer wie Hysterese
	clr		AH
	cp		AV_L,AL
	cpc		AV_H,AH
	brlo	start						; wenn kleiner, ganz zurck zu Anfang
	cli									; Interrupts aus
	sbrc	Key_Press, Taste1			; wenn Taste1 gedrckt zum Laden
	rjmp	laden
	sei									; Interrupts zulassen
	rjmp	akku_on						; und zurck
	

;*************************************************************************************************
;
; 1. Initialisierung 12 oder 24 Volt
;
;*************************************************************************************************

laden:		
	clr		Key_press					; Key_press lschen
	
	
;LCD-Strom-Korrekturwert ausrechnen
	push	UH							; UH zwischenspeichern
	ldi		AL,4						; fr die ersten 4 Stufen
	add		UH,AL						; bei UH=0 = 4*500mA = 2Amp
	ldi		AL,I_Korr					; Korrekturfaktor(I_Korr) laden
	tst		Al							; Ist AL =>0 dann weiter
	brge	j2							; bei <0 Zweierkomplement
	neg		AL							; aus -1 wird +1
j2:
	lsl		AL							; Faktor * 4
	lsl		AL

	clr		BL							; BL lschen
j1: add		BL,AL						; pro Zhler in UH, Faktor addieren
	dec		UH							; UH=UH-1
	tst		UH							; wenn UH<>0
	brne	j1							; zurck
	mov		I_Adjust,BL					; und BL in I_Adjust speichern
	pop		UH							; UH wieder vom Stack holen


;1. Strom / Startwert bei 2000mA steht in I_charge, pro 500mA mehr I_Step addieren
	ldi		AL, low(I_Charge)			; Startwert holen
	ldi		AH, high(I_Charge)			
	ldi		BL, I_Step
	
	mov		I_Charge_l,AL				; Anfangsstrom = 2 A
	mov		I_Charge_h,AH	
	
	tst		UH							; ist Flag UH (Strom) = 0 = 2 Amp
	breq	i2							; dann gleich weiter mit Laden

i1:	add		AL, BL						; sonst pro Zhler I_Step addieren
	adc		AH, ZERO					; ggf. Carry addieren
	dec		UH							; Zhler dekrementieren
	brne	i1							; in AH:AL steht nun der gewhlte Ladestrom

i2:	ldi		ZL,Low(Ladestrom)			; Max. Ladestrom im DSEG speichern
	ldi		ZH,High(Ladestrom)
	st		Z+,AH
	st		Z, AL
		
;Grenzwert fr 2. Ladestufe festlegen
	push	UL
	rcall	div8						; AH:AL / 10 = BL
	sts		grenzwert, BL				; BL im SDRAM speichern
	pop		UL


;2. Spannung 	 
	mov		V_FLAG, UL					; Spannung in Register sichern >0=12V 0=24V
	tst		UL							; ist Flag (Spannung) <> 0 dann gleich zum Laden
	brne	c1							; zum Laden
	
; Anpassung bei 24 Volt	
	lsl		FAST_V_L					; Spannung verdoppeln:
	rol		FAST_V_H					; Ladeschluspannung
	lsl		TRICKLE_V_L					; Erhaltungsladung
	rol		TRICKLE_V_H
	
	ldi		ZL,Low(Ladestrom)			; Strom halbieren
	ldi		ZH,High(Ladestrom)
	ld		UH,Z+
	ld		UL,Z
	lsr		UL
	ror		UH
	ldi		ZL,Low(Ladestrom)			; zurckschreiben
	ldi		ZH,High(Ladestrom)
	st		Z+,UH
	st		Z, UL
			
	lsr		I_Charge_H					; Strom halbieren
	ror 	I_CHARGE_L
	
	lsr		BL							; Stromlade-Grenzwert halbieren
	sts		Grenzwert, BL
	lsr		I_Adjust					; LCD-Korrekturfaktor halbieren


;*************************************************************************************************
;
; 1. Laden mit konstantem Strom bis Spannung bei 14,8 V bzw. 29,6 V
;
;*************************************************************************************************


	; "1. Stufe:" ausgeben				
c1:
	ldi		AL,0xA						; 1.Zeile 11. Stelle
	rcall	lcd_cursor
 	ldi   	ZL,Low(txtMode*2)  			; Label "Stromladen: "
 	ldi   	ZH,High(txtMode*2)
 	rcall	lcd_print		
	
	clr		Zero						; Zhler setzen, alle 256 Durchlufe ...
	clr		BL							; Additionsregister
	clr		BH							; fr Strommesswerte
	clr		CL
	clr		CH							; CH als Zhler verwenden
	sei									; Interrupts zulassen

	rcall   start_pwm					; PWM starten
	
c2:	rcall   get_current					; Strom messen, Rckgabe in AV_H:AV_L
	add		BL, AV_L					; aufaddieren
	adc		BH, AV_H
	adc		CL, Zero					; Null + Carry addieren

	mov		UL, I_CHARGE_L				; Vergleichswerte und
	mov		UH, I_CHARGE_H				
	rcall	adj_pwm						; AV_H:AV_L an Reglung bergeben
		
	dec		CH							; Zhler = Zhler -1
	tst		CH
	brne	c2							; wenn nicht 0, zurck zu c2

	mov		AV_L,BH						; Durchschnittswert aus 256 Durchlufen
	mov		AV_H,CL						; BL fllt weg = Division durch 256

	rcall	print_current				; LCD-Ausgabe Strom
	
	rcall	stop_pwm					; PWM aus 
	rcall  	get_voltage					; Spannung messen
	rcall	start_pwm					; PWM wieder ein
	
	rcall	print_voltage				; und LCD-Ausgabe Spannung
		
	cp		AV_L, FAST_V_L				; Spannung = 14,8V bzw. 29,6V
	cpc		AV_H, FAST_V_H				
 	brsh	vcf							; wenn >=, dann nchste Stufe

	clr		BL							; Additionsregister wieder auf 0
	clr		BH
	clr		CL
	clr		CH							; Zhler neu setzen

	sbrc	Key_Press, Taste1			; Wenn Taste gedrckt,
	rjmp	ende						; Programm abbrechen
		
c4:	rjmp	c2							; und zurck


;*************************************************************************************************
;
; 2. Weiterladen fr 60 min mit konstanter Spannung bei 14,8 V
;
;*************************************************************************************************
	
vcf:									
	;2. Stufe auf LCD ausgeben				
	ldi		AL,0xA						; 1.Zeile 11. Stelle
	rcall	lcd_cursor
 	ldi   	ZL,Low(txtMode*2)  			; Label 
 	ldi   	ZH,High(txtMode*2)
	adiw	ZH:ZL,8
 	rcall	lcd_print	
	
	ldi   	AL,0x49 					; Cursor von LCD auf 2. Zeile 9. Stelle
 	rcall 	lcd_cursor
	ldi   	ZL,Low(txtSonstiges*2)  	; Leerstring zur berdeckung
 	ldi   	ZH,High(txtSonstiges*2)		; der bisherigen Stromanzeige
	rcall	lcd_print	

	clr		T_Msec						; Zeit-Zhler Milli-Sekunden
	clr		T_Sec						; Zeit-Zhler Sekunden
	clr		T_Min						; Zeit-Zhler Minuten 

	clr		BL							; Additionsregister
	clr		BH							; auf 0 setzen
	clr		CL
	clr		CH							; Zhler auf 0 setzen									
	clr		ZL

vcf1:
	rcall	get_voltage					; Spannung messen, Rckgabe AV_L, AV_H
	
	add		BL, AV_L					; aufaddieren
	adc		BH, AV_H
	adc		CL, Zero					; Null + Carry addieren

	mov 	UL, FAST_V_L				; Max-Werte holen Fast-Volt (14,8 V)
	mov		UH, FAST_V_H 				; Low-und High-byte
	rcall   adj_pwm						; an Regelung bergeben
		
	dec		CH							; Zhler = Zhler -1
	tst		CH
	brne	vcf1						; wenn nicht 0, zurck zu vcf1

	mov		AV_L,BH						; Durchschnittswert aus 256 Durchlufen
	mov		AV_H,CL						; BL fllt weg = Division durch 256
	rcall	print_voltage				; 1/Sek Spannung ausgeben
	cpi		t_min,60					; wenn 60 Minuten um
	breq	vct							; weiter mit Trickle-Charge

	rcall	print_minutes				; Minutenanzeige
	
	cp		ZL,t_min					; Ist ZL gleich Minuten? d.h. nur 1 mal/Minute
	breq	vcf2						; dann weiter
	
	rcall	get_current					; wenn neu, dann Strom messen
	rcall	tab_shr						; Puffer schieben, Rckgabe Durchschnitt in AH:AL
	sts		Puffer+1,AV_L				; letzten Wert in Puffer speichern
	sts		Puffer,AV_H
	
	mov		ZL,t_min					; Zeitwert in ZL speichern
	cpi		ZL,5						; erst nach 5 Minuten zum ersten Mal
	brlo	vcf2						; Durchschnitt mit aktuellem Wert vergleichen
	
	lds		BL,Grenzwert				; Grenzwert holen
	cp		AL,BL						; Vergleich aktueller Wert (AH:AL)
	cpc		AH,Zero	   					; wenn Grenzwert grer als Durchschnitt
	brlo	vct							; dann Ende und weiter mit Trickle-Charge


vcf2:
	clr		BL							; Additionsregister wieder auf 0
	clr		BH
	clr		CL
	clr		CH							; Zhler neu setzen									
	
	sbrc	Key_Press, Taste1			; Wenn Taste gedrckt,
	rjmp	ende						; Programm abbrechen

	rjmp	vcf1						; sonst zrck


;*************************************************************************************************
;
; 3. Erhaltungsladung min mit konstanter Spannung bei 13,8 V/27,6V fr alle Ewigkeit
;
;*************************************************************************************************

vct:
	ldi		AL,0xA						; 1.Zeile 11. Stelle
	rcall	lcd_cursor	
 	ldi   	ZL,Low(txtMode*2)  			; Label
 	ldi   	ZH,High(txtMode*2)
	adiw	ZH:ZL,16
 	rcall	lcd_print		

	ldi   	AL,0x49 					; Cursor von LCD auf 2. Zeile 9. Stelle
 	rcall 	lcd_cursor
	ldi   	ZL,Low(txtSonstiges*2)  	; Leerstring zur berdeckung
 	ldi   	ZH,High(txtSonstiges*2)		; der bisherigen Stromanzeige
	adiw	ZL:ZH,8						; "Ende!" ausgeben
	rcall	lcd_print	

	clr		BL							; Additionsregister
	clr		BH							; auf 0 setzen
	clr		CL
	clr		CH

vct1:	
	rcall	get_voltage					; Spannung messen, Rckgabe AV_L, AV_H
	
	add		BL, AV_L					; aufaddieren
	adc		BH, AV_H
	adc		CL, Zero					; Null + Carry addieren

	mov 	UL, TRICKLE_V_L				; Max-Werte holen Trickle-Volt (13,8 V/27,6 V)
	mov		UH, TRICKLE_V_H 			; Low-und High-byte
	rcall   adj_pwm						; an Regelung bergeben
		
	dec		CH							; Zhler = Zhler -1
	tst		CH
	brne	vct1						; wenn nicht 0, zurck zu vct1

	mov		AV_L,BH						; Durchschnittswert aus 256 Durchlufen
	mov		AV_H,CL						; BL fllt weg = Division durch 256
	rcall	print_voltage				; Spannung ausgeben

	clr		BL							; Additionsregister wieder auf 0
	clr		BH
	clr		CL
	clr		CH							; Zhler neu setzen									

	sbrc	Key_Press, Taste1			; Wenn Taste gedrckt,
	rjmp	ende						; Programm abbrechen

	rjmp	vct1


;*************************************************************************************************
;
; 4. Ende
;
;*************************************************************************************************
ende:
	rcall	stop_pwm
	rjmp	poweron

	

;*************************************************************************************************
;
; Timer0-Interrupt
;
;*************************************************************************************************


TIM0_OVF:								; Zhlt bis 10 Minuten
	push	UL							 
	push	UH
	push	BL
	push	ZH
	push	ZL
	in		R21,SREG					; Inhalt von SREG zwischenspeichern
	mov		UL,Key_old					
	in		Key_old, Key_Port			
	eor		UL, Key_Old					
	com		Key_old
	mov		UH, Key_State
	or		Key_State, UL
	and		UL, Key_Old
	eor		Key_State,UL
	and 	UH, UL
	or		Key_Press, UH

	inc 	T_Msec						; Milli-Sek. erhhen
	cpi		T_Msec,16					; sind es 15
	brlo	t0_end						; wenn nein, dann Schleife
	clr		T_Msec						; Millisekunden auf 0
	inc		T_Sec						; und Sekunden erhhen (1049 ms)
	
	;Vergleich aktueller Strom mit Sollstrom
	ldi		ZL, low(Ladestrom)
	ldi		ZH, high(Ladestrom)	
	ld		UH,Z+
	ld		UL,Z
	cp		I_Charge_l, Ul
	cpc		I_Charge_h, UH
	brge	TO_0						; wenn >= dann zu TO_0
	ldi		BL,6
	add		I_Charge_l,BL				; sonst pro Sekunde um ca. 100 mA
	adc		I_Charge_h,Zero				; erhhen
	rjmp	TO_1
TO_0:
	mov		I_Charge_l,UL				; falls UL und UH grer waren
	mov		I_Charge_h,UH				; hier auf gleich setzen
	
TO_1:
	cpi		T_Sec,58					; sind es 60 Sekunden?
	brlo	t0_end						; wenn nicht, dann Schleife
	clr		T_Sec						; Sekunden auf 0
	inc		T_Min						; und Minute erhhen (59752 ms)
T0_end:
	out		SREG, R21					; Inhalt von SREG wieder herstellen
	pop		ZL
	pop		ZH
	pop		BL
	pop		UH
	pop		UL
	reti


;*************************************************************************************************
;
; ADC-Handler, Eingabe: ADMUX = UL Rckgabe: AV_H:AV_l
;
;*************************************************************************************************

get_current:
	ldi		UL, ADC_Diff_Channel
	rjmp	get_measure			

get_voltage:
	ldi		UL, ADC_V_Channel

get_measure:
	cli									; Whrend Messung, Interrupts verhindern
	out		ADMUX, UL					; ADC Channel setzen
	rcall	wait1ms						; warte 1ms
	clr		AV_L
	clr		AV_H						; Additionsregister lschen
	
	ldi		UL, (1<< ADEN) 				; 
	out		ADCSR,UL
	
	sbi		ADCSR,ADSC					; 1. Start = Dummymessung
m_dummy:								
	sbis	ADCSR,ADIF
	rjmp	m_dummy

	ldi		UH,32 						; Schleifenzhler auf 32

m_start:
	sbi		ADCSR, ADSC					; 2. Start

m_wait:
	sbic	ADCSR, ADSC
	rjmp	m_wait						; Warte bis Messung erfolgt ist (ADSC wird 0)

	in		UL, ADCL					; Messungen zwischenspeichern
	add		AV_L, UL					; und
	in		UL, ADCH
	adc		AV_H, UL					; Werte aufaddieren

	dec		UH
	brne	m_start						; wiederholen bis Zhler=0
	
	ldi		UH, 5 						; Zhler fr's Schieben, 5 mal schieben = geteilt durch 32
m_avg:	
	lsr		AV_H
	ror		AV_L						; 
	dec		UH							; Zhler=Zhler-1
	brne	m_avg						;	 
	
	cbi		ADCSR,ADEN
	sei									; Interrupts wieder zulassen
	ret									; 
	

;*************************************************************************************************
;
; PWM-Handler: Eingabe Spannung in AV_H:AV_l und Vergleichswert in UH:UL
;
;*************************************************************************************************

;PWM-Regelung
adj_pwm:
	cp		AV_L,UL						; Vergleich Register, Register
	cpc		AV_H,UH						; Vergleich Register, Register + Carry
	breq	f_end						; wenn gleich, dann Ende
	cp		AV_L,UL						; nochmal vergleichen
	cpc		AV_H,UH			
	brsh	f_dec						; springe bei grer zu dec	

f_inc:									; ansonsten erhhen
	in		UL, OCR1A					; aktuellen Wert einlesen
	inc		UL							; UL=UL+1
	cpi		UL, 159						; ist UL>= OCR1C
	brsh	f_end						; dann Ende
	out		OCR1A, UL					; sonst UL nach OCR1A
	rjmp 	f_end					

f_dec:	
	in		UL, OCR1A					; aktuellen Wert einlesen
	dec		UL							; OCR1A > 0 ?
	cpi		UL,1
	brlo	f_end						; wenn Carry 0 dann vorher zu ende
	out		OCR1A, UL

f_end:
	ret


; Start and Stop PWM
start_PWM:
	sbi		PORTB,PWM_PIN
	sbi		DDRB, PWM_PIN	
	ldi 	UL, (1<< COM1A1)|(1 << PWM1A)
	out 	TCCR1A, UL						
	tst		V_Flag						; wenn V_FLAG = 0 dann 24V
	breq	p1				

	; bei 12 V 50 khz
	ldi 	UL,   (1 << CS12)			; bei 12 = 64 Mhz/ 8/159=50 khz
	out 	TCCR1B, UL					
	ret						
												
p1: ; bei 24V 25 khz	
	ldi 	UL,   (1 << CS12)|(1<<CS10)	; 12+10  = 64 Mhz/16/159=25 khz
	out 	TCCR1B, UL					
	ret						
												
												
stop_PWM:
	cbi		PORTB, PWM_PIN				
	sbi		DDRB,  PWM_PIN			
	ldi		UL, (0 << PWM1A)
	out		TCCR1A, UL					; PWM1A auf null setzen
	ret


;*************************************************************************************************
;
; vier Werte in Puffer nach rechts schieben, vorher aufaddieren und Durchschnitt berechnen
; Rckgabewert in AH:AL
;
;*************************************************************************************************

tab_shr:
	lds		UL, Puffer+7		; 1.Wert (Low-Byte) lesen
	mov		AL,UL				; in AL speichern
	lds		UL, Puffer+6		; 1. Wert(high-Byte lesen)
	mov		AH,UL				; in AH speichern
	
	lds		UL, Puffer+5		; 2.Wert (low) lesen
	sts		Puffer+7,UL			; und an 1. Stelle (low) speichern
	add		AL,UL				; zu AL addieren
	
	lds		UL, Puffer+4		; 2.Wert (high) lesen
	sts		Puffer+6,UL			; und an 1. Stelle (high) speichern
	adc		AH,UL				; zu AH addieren

	lds		UL, Puffer+3		; usw.
	sts		Puffer+5,UL
	add		AL,UL

	lds		UL, Puffer+2
	sts		Puffer+4,UL
	adc		AH,UL

	lds		UL, Puffer+1
	sts		Puffer+3,UL
	add		AL,UL
	
	lds		UL, Puffer
	sts		Puffer+2,UL
	adc		AH,UL

	lsr		AH					; Durchschnitt der 4
	ror		AL					; Werte durch 2 mal
	lsr		AH					; rechtsschieben
	ror		AL
	ret


;*******************[ Warteschleifen ]*******************************************************

wait1s: ;1002,54
	push		UL
	ldi			UL,10
n1:	rcall		wait100ms
	dec			UL
	brne		n1
	pop			UL
	ret

wait100ms:; 100,54ms
	push		UL
	ldi			UL,10
n2:	rcall		wait10ms
	dec			UL
	brne		n2
	pop			UL
	ret


wait5ms:; 5,03ms
	push		UL
	ldi			UL,5
	rjmp		n3
wait10ms: ;10,04ms
	push		UL
	ldi			UL,10
n3:	rcall		wait1ms
	dec			UL
	brne		n3
	pop			UL
	ret

wait1ms:
	push		UL
	ldi			UL,9
n4:	rcall		wait100us
	dec			UL
	brne		n4
	pop			UL
	ret

wait50us:	;48us 
	push		UL
	ldi			UL,14
	rjmp		n5
wait100us:;	100us
	push		UL
	ldi			UL,32
n5:	dec			UL
	brne		n5
	pop			UL
	ret


;********************************************************************************
; Spannung auf LCD ausgeben
; Werte AV_H:AV_L
; 2. Zeile 1. Stelle
;**********************************************************************************

Print_voltage:										
 
	ldi   	AL,0x40  						; Cursor von LCD auf 2. Zeile 1 Stelle
 	rcall 	lcd_cursor
 
 	mov   	AL, AV_L						; Messung speichern
 	mov   	AH, AV_H
 	add		AL, AV_L
 	adc		AH, AV_H
 	add		AL, AV_L
 	adc		AH, AV_H						; jetzt ist der dreifache Wert in AH:AL
	
	rcall 	bin2ascii						; Wert in AH:AL in Ascii umrechnen
 	
	push 	AL								; zuerst 1er Stelle sichern
 	push	AH                            	; 10er Stelle
 	push	BL								; 100er Stelle
 	mov   	AL,BH							; in BH steht 1000er Stelle
 	rcall	lcd_data						; ausgeben
 	pop 	AL								; 100er Stelle holen
 	rcall 	lcd_data                     	; ausgeben
 	ldi   	AL,','							; ein Komma zwischensetzen
 	rcall 	lcd_data                     	; ausgeben
 	pop   	AL		                        ; 10er Stelle holen
 	rcall 	lcd_data                     	; ausgeben
 	pop   	AL								; 1er Stelle holen
 	rcall 	lcd_data                     	; ausgeben
 	ldi   	AL,' '							; Leerstelle dranhngen
 	rcall 	lcd_data                     	; ausgeben
 	ldi   	AL,'V'							; V fr Volt dranhngen
 	rcall 	lcd_data                     	; ausgeben
	ret 
 

;**********************************************************************************
; Strom auf LCD ausgeben 2A = 124 Steps = ca. 124 * 16 = 1984
; Werte AV_H:AV_L
;**********************************************************************************

Print_Current:										
 
	ldi   	AL,0x49 						; Cursor von LCD auf 2. Zeile 8. Stelle
 	rcall 	lcd_cursor
  	
 	mov  	AL, AV_L						; Messung speichern
	mov   	AH, AV_H
	
pc2:	
	lsl		AL								; 4 mal linksschieben = mal 16
	rol		AH
	lsl		AL
	rol		AH								; 
	lsl		AL
	rol		AH								; 
	lsl		AL
	rol		AH								; 
	
	ldi		UL, I_Korr						; Korrekturfaktor (signed!) nach UL
	tst		UL								; >0 oder <0 ?
	brge	pc3								; wenn UL >=0 weiter bei pc3
	sub		AL, I_Adjust					; Korrekturwert (unsigned!) subtrahieren
	sbc		AH, Zero						; ggf. Carry-Flag
	rjmp	pc4								; weiter mit Ausgabe

pc3:	
	add		AL, I_Adjust					; Korrekturwert (unsigned!) addieren
	adc		AH, Zero						; ggf. Carry-Flag addieren

pc4:										; Ausgabe
	;in AV_H:AV_L steht jetzt der Wert in mA
	rcall 	bin2ascii						; Wert in AH:AL in Ascii umrechnen
 											; 1er Stelle in Al wird verworfen
 	push	AH                            	; 10er Stelle 
 	push	BL								; 100er Stelle
 	ldi		AL, '0'							; 1000 Stelle immer mit 0 ausgeben
	rcall	lcd_data						; ausgeben
	mov   	AL,BH							; in BH steht 1000er Stelle
 	rcall	lcd_data						; als 100 ausgeben
 	ldi   	AL,','							; ein Komma zwischensetzen
 	rcall 	lcd_data                     	; ausgeben
	pop 	AL								; 100er Stelle vom Stack holen nach AL
 	rcall 	lcd_data                     	; ausgeben
 	pop   	AL		                        ; 10er Stelle vom Stack holen
 	rcall 	lcd_data                     	; ausgeben
 	ldi   	AL,' '							; Leerstelle dranhngen
 	rcall 	lcd_data                     	; ausgeben
 	ldi   	AL,'A'							; A fr Ampere dranhngen
 	rcall 	lcd_data                     	; ausgeben
	ret


;**********************************************************************************
; Zeit auf LCD ausgeben
; Werte AV_H:AV_L
;**********************************************************************************

Print_Minutes:
 
 	ldi   	AL,0x4A 						;Cursor von LCD auf 2. Zeile 10. Stelle
 	rcall 	lcd_cursor
  	
 	ldi  	AL, 60
 	sub   	AL, t_min
	clr		AH	

	;in AL steht jetzt die verbeibende Zeit
	rcall 	bin2ascii						; Wert in AH:AL in Ascii umrechnen
 											; 1er Stelle in Al wird verworfen
 	push	AL
 	mov		AL, AH							; 10 Stelle ausgeben
	rcall	lcd_data						; ausgeben
	pop 	AL								; 1er Stelle holen
 	rcall 	lcd_data                     	; ausgeben
 	ldi   	AL, ' '		                    ; Leerzeichen
 	rcall 	lcd_data                     	; ausgeben
 	ldi   	AL,'m'							; und min
 	rcall 	lcd_data                     	; anfgen
 	ldi   	AL,'i'							; 
 	rcall 	lcd_data                     	
 	ldi   	AL,'n'							; 
 	rcall 	lcd_data                     	
	ret


;*************************************************************************
; Binrwert in Ascii-Code wandeln 
; Eingabe: AH:AL, Rckgabe in  CL, BH:BL, AH:AL = 5 digits (ASCII)
;*************************************************************************

bin2ascii:
	ldi  	CL, -1 + '0'
bcd1: 
	inc  	CL
    subi 	AL, low(10000)
    sbci 	AH, high(10000)
    brcc 	bcd1
    ldi  	BH, 10 + '0'

bcd2:
	dec		BH
    subi 	AL, low(-1000)
    sbci 	AH, high(-1000)
    brcs 	bcd2
    ldi  	BL, -1 + '0'

bcd3:
	inc  	BL
    subi 	AL, low(100)
    sbci 	AH, high(100)
    brcc 	bcd3
    ldi  	AH, 10 + '0'

bcd4:
	dec  	AH
    subi 	AL, -10
    brcs 	bcd4
    subi 	AL, -'0'

	ret



;***************************************************************************
;*
;* LCD-Routinen  4-Bit
;*                                             	
;* PIN_LCD_Enable	= PB5									
;* PIN_LCD_RS		= PB6									
;* PIN_LCD_D4		= PA4									
;* PIN_LCD_D5		= PA5									
;* PIN_LCD_D6 		= PA6									
;* PIN_LCD_D7		= PA7									
;*																
;****************************************************************************


;sendet Datenbyte an das LCD
lcd_data:

	sbi		PortB, PIN_LCD_RS
	push	AL		          	   		;"Sicherungskopie" 2.Nibble
    andi  	AL, 0b11110000         		; unteres Nibble auf Null setzen
	out		PORTA,AL
	rcall 	lcd_enable             		; Enable-Routine aufrufen
    pop   	AL                	   		; 2. Nibble
    swap  	AL				   	   		; untere 4 bits (Nibble) nach oben	
  	andi  	AL, 0b11110000         		; untere Hlfte auf Null setzen
	out		PORTA,AL
	rcall 	lcd_enable             		; Enable-Routine aufrufen
    rcall 	wait50us              		; warte
    ret                            		; zurck zum Hauptprogramm
	

 ;sendet einen Befehl an das LCD
lcd_command:                            ; wie lcd_data, nur ohne RS zu setzen
    		
	cbi		PORTB, PIN_LCD_RS
	push	AL
    andi	AL, 0b11110000
	out		PORTA,AL
	rcall	lcd_enable
    pop		AL
	swap	AL
	andi	AL, 0b11110000
	out		PORTA,AL
	rcall	lcd_enable
    rcall	wait50us
    ret


 ;erzeugt den Enable-Puls
lcd_enable:
    
 	sbi 	PORTB, PIN_LCD_Enable    	; Enable high
    nop                          		; 3 Taktzyklen warten
	nop
	nop
	cbi		PORTB, PIN_LCD_Enable     	; Enable wieder low
	ret                          		; und wieder zurck

 
 ;Initialisierung
lcd_init:
    rcall  	wait1s

	ldi	    AL, 0b0011000         		;
    out		PORTB, AL              		;
	rcall	lcd_enable             		; 1. muss 3mal hintereinander gesendet
    rcall	wait5ms						;    werden zur Initialisierung
    rcall	lcd_enable             		; 2.
    rcall	wait5ms
    rcall	lcd_enable             		; und 3!
    rcall	wait5ms
	
	ldi		AL,	0b00100000				; 4bit Modus einstellen
	out		PORTB,AL
	rcall	lcd_enable
	rcall	Wait5ms	

	ldi	    AL, 0b00101000         		; 4bit-Modus, 2 Zeilen
    rcall	lcd_command					; Befehl senden
	ldi		AL, 0b00001100         		; LCD on
    rcall	lcd_command					; Befehl senden
	ldi		AL, 0b00000100         		; Entry mode
    rcall	lcd_command
    ret

 ;Sendet den Befehl zur Lschung des Displays
lcd_clear:
    ldi		AL, 0b00000001         		; Display lschen
    rcall	lcd_command
    rcall	Wait5ms
    ret
 ;Setzt Cursor-Position auf Anfang
lcd_home:
	ldi		AL, 0b00000010         		; Cursor an Anfang
   	rcall	lcd_command
   	rcall	wait5ms
   	ret	

; Erwartet die Position in AL			
lcd_cursor:
	sbr		AL, 0b10000000         		; Bit 7 setzen
	rcall	lcd_command
	rcall	wait5ms
    ret

lcd_print:
	lpm									; Laden
	tst		R0							; ist Zeichen =0
	breq	lcd_pend					; gehe zu end
	mov		AL,R0						; sonst Zeichen nach AL
	rcall	lcd_data					; Ausgabe
	adiw	ZL,1              			; Zeiger in ZL = ZL + 1(word)
	rjmp	lcd_print				
lcd_pend:
 	ret


;*************************************************************************
; Dividiert 16Bit-Zahl durch 8Bit-Zahl
; Eingabe:  AH:AL, Rckgabe in  BH:BL, CL = Divisor
; Hilsregister UL, UH
;*************************************************************************


div8:
	
	ldi		CL,10			; Divisor
	ldi		BL,1			; zhlt von 1 an
	clr		UL				; Hilfs-
	clr		BH				; und Ergebisregister lschen

div8a:
	clc						; lscht Carry
	rol		AL				; hchstes Bit 
	rol		AH				; ins
	rol		UL				; Hilfsregister rollen
	brcs	div8b			; wenn berlauf vorhanden zu div8b
	cp		UL,CL			; Ergebnis 1 oder 0?	
	brcs	div8c			; wenn kleiner keine Subtraktion

div8b:
	sub		UL,Cl			; subtrahiere Divisor
	sec						; setze Carry
	rjmp	div8d			; schiebe das Ergebnis

div8c:
	clc						; lscht Carry

div8d:
	rol		BL				; rotiere Carry ins Ergebnis
	rol		BH
	brcc	div8a			; mach weiter, solange Carry = 0

	cpi		UL,5			; Rest steht in UL, ist UL<5,
	brlo	div8e			; dann weiter
	inc		BL				; sonst 1 addieren
	adc		BH, Zero

div8e:
	ret




