

;*********** [ Blei-Akku-Lader fr 12 und 24 Volt ] ********************
;
;   Datei:		SLA.INC
;   Autor:		Hanno Grff
;   Version:	1.3
;   Datum:		August 2007
;
;   Compiler:	Atmel AVR Assembler
;
;
;**********************************************************************
;
;   Vorabgleich von AREF mit Netzgert und 14,4 Volt an den
;	Batterieklemmen ergibt ca. AREF von 3,66 V. Besser ist es, 
;	whrend des Ladens unter Last Voltmeter an Batterie klemmen
;	und LCD-Anzeige abgleichen, ergibt AREF von ca. 3,58 bis 3,60 V
;
;	I_Charge hngt in erster Linie von Rshunt=R9/91 ab !!!
;	Bsp: Ishunt = 0,11 ohm => I_Charge = 150
;		 Ishunt = 0,10 ohm => I_Charge = 136,5
;		 Ishunt = 0,09 ohm => I_Charge = 123
;
;
;	Folgende Rechnungen gelten fr 0,1 ohm
;	--------------------------------------------
;
;	AREF		=   3660 mV (messen und einstellen)
;               =   3600 mV (unter Last abgleichen)
;
;
;   ADC und Spannung und am ADC-Diff-Eingang:
;
;	Step        =   (20 * Ishunt * R7/R10 * 1024 * Rshunt) / ((R7/R10 + R8/R11) * AREF)
;				=   (20 * 2000mA * 3 kOhm * 1024 * 0,1   ) / ((3 kOhm +22 kOhm) * 3600mV)
;				=   136 
;
;
;	Steps		=   3600mV / 1024 Steps = 3,52 mV/Step
;
;	VADC		=   Step * 3,52mV
;	
;			;	
;	IMax 		=   AREF / 0,24
;					3600mV / 0,24 = 15000 mA
;
;					
;	I_Charge  	=	2000 mA <=>  481 mV / 3,52 <=> Step = 136
;					3000 mA	<=>  721 mV / 3,52 <=> Step = 201
;					4000 mA	<=>  961 mV / 3,52 <=> Step = 273
;
;
;                   Ladeschluspannung * R7_10 * AD_resolution
;   V_Charge_12h  = __________________________________________
;                   (R8_11 + R7_10) * V_ref_voltage
;
;		    	  =	1,48V * 3000 * 1024 / (22000+3000) * 3,66V
;		    	  
;				  = 495
;
;
;	V_Charge_12l  =	1,38V * 3000 * 1023 / 25000 * 3,66V
;				  
;				  = 463
;				

;	I_Charge = Step bei 2000mA = 136     siehe oben, im Idealfall bei Rshunt von genau 0,1 Ohm

;	I_Step 	 = Step bei 500 mA =  34  

;	Grenzwert bei 		200 mA =   13,6  siehe Tabelle Grenzwerte in SLA.ASM
;						300 mA =   20,4
;						usw.

; 	I_Korr   = 125 - I_Charge            bei I-Charge von 125 entspricht die LCD-Anzeige genau 2,00 Amp (125*16=2000)
;										 Bsp: bei I_Charge von 136 ist I-Korr: 125 - 136 = -11


; 	Folgende Werte sind experimentell festzulegen:
; 
;	Fast Charge Voltage Limit
	.equ    V_Charge_fast    =  495     ; Siehe obige Formel

;	Trickle Charge Voltage Limit
	.equ    V_Charge_trickle =  460   


;	Max Charge Current
	;.equ	I_Charge	= 136		; bei 12 Volt Startwert entspricht: 2000 mA (Idealfall bei genau 0,1 Ohnm)
	;.equ	I_Step 		= 34
	.equ	I_Charge	= 124		; bei 12 Volt Startwert entspricht: 2000 mA
	.equ	I_Step		= 31		; bei 12 Volt Additionswert entspricht: 500 mA = I_Charge / 4
	
;	Strom-Korrekturfaktor fr LCD-Anzeige: bei I_Charge von 125 * 16 = 2000 mA
; 	im Normalfall 125 - I_Charge = Korrekturfaktor
; 	sonst nachjustieren
	.equ	I_Korr		= 1


